/*
 * cmdu_validate.c
 * WiFi MultiAP CMDU validation functions
 *
 * Copyright (C) 2020-2022 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <string.h>
#include <easymesh.h>
#include <1905_tlvs.h>
#include <cmdu.h>
#include <map_module.h>
#include "cmdu_validate.h"
#include "debug.h"

bool validate_channel_scan_report(struct cmdu_buff *cmdu, struct tlv *tv_tsp[][TLV_MAXNUM],
				  struct tlv *tv_scan[], int *num,
				  uint8_t map_profile)
{
	int i, j;
	int ret = 0;

	struct tlv_policy d_policy_scan[] = {
		[0] = {
			.type = MAP_TLV_CHANNEL_SCAN_RES,
			.present = TLV_PRESENT_MORE,
			.minlen = 9
		}
	};

	ret = map_cmdu_parse_tlvs(cmdu, tv_tsp, 2, map_profile);

	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		if (!tv_tsp[0][0])
			fprintf(stderr, "%s: No TIMESTAMP_TLV received!\n", __func__);

		return false;
	}

	ret = cmdu_parse_tlv_single(cmdu, tv_scan, d_policy_scan, num);
	if (ret) {
		fprintf(stderr, "%s: cmdu_parse_tlv_single failed\n", __func__);
		return false;
	}


	if (!tv_scan[0]) {
		fprintf(stderr, "%s: No RESULT_TLV received!\n", __func__);
		return false;
	}

	/* Timestamp TLV */
	if (tv_tsp[0][0]) {
		int offset = 0;
		int tsp_len = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv_tsp[0][0]);

		t = (uint8_t *)tv_tsp[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		tsp_len = t[offset];
		offset += 1;
		if ((offset + tsp_len) > tlv_len)
			return false;
	}

	/* Channel Scan Result TLV */
	for (i = 0; i < *num; i++) {
		int offset = 0;
		uint8_t status = 0x00;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv_scan[i]);
		int ts_len = 0;
		uint16_t num_nb = 0;

		t = (uint8_t *)tv_scan[i]->data;
		if (!t)
			return false;

		/* offset within provided minlen(9) */
		offset += 6;	/* radio id */
		offset += 1;	/* opclass */
		offset += 1;	/* channel */
		status = t[offset];
		offset += 1;	/* status */

		if (status == CH_SCAN_STATUS_SUCCESS) {

			if ((offset + 1) > tlv_len)
				return false;

			ts_len = t[offset];
			offset += 1;	/* timestamp len */
			if ((offset + ts_len + 4) > tlv_len)
				return false;

			offset += ts_len;	/* timestamp */
			offset += 1;		/* utilization */
			offset += 1;		/* noise */
			num_nb = BUF_GET_BE16(t[offset]);
			offset += 2;		/* num neightbors */

			for (j = 0; j < num_nb; j++) {
				int ssid_len = 0;
				int bw_len = 0;
				uint8_t info = 0x00;

				if ((offset + 7) > tlv_len)
					return false;

				offset += 6;	/* bssid */
				ssid_len = t[offset];
				offset += 1;	/* ssid len */
				if ((offset + ssid_len + 2) > tlv_len)
					return false;

				offset += ssid_len;	/* ssid */
				offset += 1;		/* rcpi */
				bw_len = t[offset];
				offset += 1;		/* bw len */
				if ((offset + bw_len + 1) > tlv_len)
					return false;

				offset += bw_len;	/* bandwidth */
				info = t[offset];
				offset += 1;		/* info */

				if (info & CH_SCAN_RESULT_BSSLOAD_PRESENT) {
					if ((offset + 3) > tlv_len)
						return false;

					offset += 1;	/* channel utilization */
					offset += 2;	/* station count */
				}
			}

			if ((offset + 5) > tlv_len)
				return false;

			offset += 4;	/* total scan duration */
			offset += 1;	/* scan type */
		}
	}

	return true;
}

bool validate_topology_response(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
				size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);

	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		if (!tlvs[0][0])
			fprintf(stderr, "%s: TLV DEVICE_INFORMATION_TYPE is missing\n", __func__);

		if (!tlvs[7][0])
			fprintf(stderr, "%s: TLV AP_OPERATIONAL_BSS is missing\n", __func__);

		if (!tlvs[9][0])
			fprintf(stderr, "%s: TLV MULTIAP_PROFILE is missing\n", __func__);

		return false;
	}

	/* AP Operational BSS TLV */
	if (tlvs[7][0]) {
		int i, j;
		int offset = 0;
		int num_radio = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[7][0]);

		t = (uint8_t *)tlvs[7][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen (1) */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			int num_bss = 0;

			if ((offset + 6 + 1) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			num_bss = t[offset];
			offset += 1;	/* num bss */

			for (j = 0; j < num_bss; j++) {
				int ssid_len = 0;

				if ((offset + 6 + 1) > tlv_len)
					return false;

				offset += 6;	/* bssid */
				ssid_len = t[offset];
				offset += 1;	/* ssid len */

				if ((offset + ssid_len) > tlv_len)
					return false;

				offset += ssid_len;	/* ssid */
			}
		}
	}

	/* Associated Client TLV */
	if (tlvs[8][0]) {
		int i;
		int offset = 0;
		int num_bss = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[8][0]);

		t = (uint8_t *)tlvs[8][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen (1) */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		for (i = 0; i < num_bss; i++) {
			uint16_t num_cl = 0;

			if ((offset + 6 + 2) > tlv_len)
				return false;

			offset += 6;	/* bssid */
			num_cl = BUF_GET_BE16(t[offset]);
			offset += 2;	/* num client */

			if ((offset + (num_cl * 8)) > tlv_len)
				return false;

			/* client macaddr + conntime */
			offset += (num_cl * 8);
		}
	}

#if (EASYMESH_VERSION > 2)
	/* MAP_TLV_BSS_CONFIGURATION_REPORT */
	if (tlvs[10][0]) {
		struct tlv_bss_configuration_report *tlv;
		uint8_t *tv_data;
		uint16_t tlv_len = 0;
		int i, offset = 0;

		tlv_len = tlv_length(tlvs[10][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_bss_configuration_report *)tlvs[10][0]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/* num_radio (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_radio; i++) {
			uint8_t num_bss = 0;
			int j;

			/* macaddr (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			offset += 6;
			/* num_bss (1 byte) */
			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_bss, &tv_data[offset], 1);

			offset += 1;

			for (j = 0; j < num_bss; j++) {
				uint8_t ssidlen = 0;

				/* macaddr (6 bytes) */
				if (offset + 6 > tlv_len)
					return false;

				offset += 6;

				/* flag (1 bytes) */
				if (offset + 1 > tlv_len)
					return false;

				offset += 1;

				/* reserved (1 bytes) */
				if (offset + 1 > tlv_len)
					return false;

				offset += 1;

				/* ssidlen (1 byte) */
				if (offset + 1 > tlv_len)
					return false;

				memcpy(&ssidlen, &tv_data[offset], 1);
				offset += 1;

				/* ssid (ssidlen bytes) */
				if (offset + ssidlen > tlv_len)
					return false;

				offset += ssidlen;
			}
		}
	}
#endif
	return true;
}

bool validate_ap_metric_resp(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
			     size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);
	int index = 0;

	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		if (!tlvs[0][0] || !tlvs[3][0])
			fprintf(stderr, "%s: required TLV missing: "
				"AP_METRICS_TLV or AP_EXTENDED_METRICS_TLV\n",
				__func__);

		return false;
	}

	/* AP Metrics TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[0][index]) {
		int offset = 0;
		uint8_t *t = NULL;
		uint8_t esp_ac = 0x00;
		uint16_t tlv_len = tlv_length(tlvs[0][index]);

		t = (uint8_t *)tlvs[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(13) */
		offset += 6;	/* bssid */
		offset += 1;	/* channel util */
		offset += 2;	/* num station */
		esp_ac = t[offset];
		offset += 1;	/* ESP */
		offset += 3;	/* ESP_BE */

		if (esp_ac & ESP_AC_BK) {
			if ((offset + 3) > tlv_len)
				return false;

			offset += 3;	/* ESP_BK */
		}

		if (esp_ac & ESP_AC_VO) {
			if ((offset + 3) > tlv_len)
				return false;

			offset += 3;	/* ESP_VO */
		}

		if (esp_ac & ESP_AC_VI) {
			if ((offset + 3) > tlv_len)
				return false;

			offset += 3;	/* ESP_VI */
		}

		index++;
	}

	/* Associated STA Link Metrics TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[2][index]) {
		int offset = 0;
		int num_bss = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[2][index]);

		t = (uint8_t *)tlvs[2][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* sta mac */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		if ((offset + 19 * num_bss) > tlv_len)
			return false;

		index++;
	}

	/* Associcated STA Extended Link Metrics TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[5][index]) {
		int offset = 0;
		int num_bss = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[5][index]);

		t = (uint8_t *)tlvs[5][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* sta mac */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		if ((offset + 22 * num_bss) > tlv_len)
			return false;

		index++;
	}

	return true;
}

bool validate_ap_cap_report(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
			    size_t tlvs_size, uint8_t map_profile)
{
	int i, j, k, ret = 0;
	int index = 0;

	ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);
	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		return false;
	}

	/* AP Radio Basic Capability TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[1][index]) {
		int offset = 0;
		int num_op = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[1][index]);

		t = (uint8_t *)tlvs[1][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(8) */
		offset += 6;	/* radio id */
		offset += 1;	/* max bss */
		num_op = t[offset];
		offset += 1;	/* num opclass */

		for (i = 0; i < num_op; i++) {
			int num_nonop_ch = 0;

			if ((offset + 3) > tlv_len)
				return false;

			offset += 1;	/* class id */
			offset += 1;	/* max txpower */
			num_nonop_ch = t[offset];
			offset += 1;	/* num nonop channel */

			if ((offset + num_nonop_ch) > tlv_len)
				return false;

			offset += num_nonop_ch;
		}

		index++;
	}

	/* AP HE Capability TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[4][index]) {
		int offset = 0;
		int he_len = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[4][index]);

		t = (uint8_t *)tlvs[4][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(9) */
		offset += 6;	/* radio id */
		he_len = t[offset];
		offset += 1;	/* HE len */

		if ((offset + he_len + 2) > tlv_len)
			return false;

		offset += he_len;
		offset += 2;

		index++;
	}

	/* Channel Scan Capability TLV */
	if (tlvs[5][0]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_radio = 0;
		int num_op = 0;
		int num_ch = 0;
		uint16_t tlv_len = tlv_length(tlvs[5][0]);

		t = (uint8_t *)tlvs[5][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			if ((offset + 12) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			offset += 1;	/* cap */
			offset += 4;	/* min scan interval */
			num_op = t[offset];
			offset += 1;	/* num opclass */

			for (j = 0; j < num_op; j++) {
				if ((offset + 2) > tlv_len)
					return false;

				offset += 1;	/* class id */
				num_ch = t[offset];
				offset += 1;	/* num channel */

				if ((offset + num_ch) > tlv_len)
					return false;

				offset += num_ch;
			}
		}
	}

	/* CAC Capability TLV */
	if (tlvs[6][0]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_radio = 0;
		int num_cac = 0;
		int num_op = 0;
		int num_ch = 0;
		uint16_t tlv_len = tlv_length(tlvs[6][0]);

		t = (uint8_t *)tlvs[6][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(3) */
		offset += 2;	/* country code */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			if ((offset + 7) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			num_cac = t[offset];
			offset += 1;	/* num cac */

			for (j = 0; j < num_cac; j++) {
				if ((offset + 5) > tlv_len)
					return false;

				offset += 1;	/* cac method */
				offset += 3;	/* duration */
				num_op = t[offset];
				offset += 1;	/* num opclass */

				for (k = 0; k < num_op; k++) {
					if ((offset + 2) > tlv_len)
						return false;

					offset += 1;	/* class id */
					num_ch = t[offset];
					offset += 1;	/* num channel */

					if ((offset + num_ch) > tlv_len)
						return false;

					offset += num_ch;
				}
			}
		}
	}

	return true;
}

bool validate_bksta_cap_report(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
			       size_t tlvs_size, uint8_t map_profile)
{
	int index = 0;
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		return false;
	}

	/* Backhaul STA Radio Capability TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[0][index]) {
		int offset = 0;
		uint8_t *t = NULL;
		uint8_t mac_include = 0x00;
		uint16_t tlv_len = tlv_length(tlvs[0][index]);

		t = (uint8_t *)tlvs[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* radio id */
		mac_include = t[offset];
		offset += 1;	/* macaddr include */

		if (mac_include & BSTA_MACADDRESS_INCLUDED) {
			if ((offset + 6) > tlv_len)
				return false;

			offset += 6;	/* bsta macaddress */
		}

		index++;
	}

	return true;
}

bool validate_operating_channel_report(struct cmdu_buff *cmdu,
				       struct tlv *tlvs[][TLV_MAXNUM], size_t tlvs_size,
				       uint8_t map_profile)
{
	int index = 0;
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		return false;
	}

	/* Operating Channel Report TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[0][index]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_opclass = 0;
		uint16_t tlv_len = tlv_length(tlvs[0][index]);

		t = (uint8_t *)tlvs[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(8) */
		offset += 6;	/* radio id */
		num_opclass = t[offset];
		offset += 1;	/* num opclass */

		if ((offset + num_opclass * 2 + 1) > tlv_len)
			return false;

		index++;
	}

	return true;
}

bool validate_autoconfig_response(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
				  size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		return false;
	}

	/* Supported Service TLV */
	if (tlvs[2][0]) {
		int offset = 0;
		int num_serv = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[2][0]);

		t = (uint8_t *)tlvs[2][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		num_serv = t[offset];
		offset += 1;	/* num supported service */
		if ((offset + num_serv) > tlv_len)
			return false;

	}

	return true;
}

bool validate_topology_notification(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
				    size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		return false;
	}


	return true;
}

bool validate_disassoc_stats(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
			     size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		if (!tlvs[0][0] || !tlvs[1][0] || !tlvs[2][0])
			fprintf(stderr, "%s: required tlv missing\n", __func__);

		return false;
	}

	return true;
}

bool validate_client_cap_report(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM],
				size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		if (!tlvs[0][0])
			fprintf(stderr, "%s: Client-Info TLV missing\n", __func__);

		if (!tlvs[1][0])
			fprintf(stderr, "%s: Cap-Report TLV missing\n", __func__);

		return false;
	}

	return true;
}

bool validate_unassoc_sta_link_metric_resp(struct cmdu_buff *cmdu,
					   struct tlv *tlvs[][TLV_MAXNUM],
					   size_t tlvs_size, uint8_t map_profile)
{
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);

	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));

		if (!tlvs[0][0])
			fprintf(stderr, "%s: required tlv missing\n", __func__);

		return false;
	}

	/* Unassociated STA Link Metrics Response */
	int offset = 0;
	int num_sta = 0;
	uint8_t *t = NULL;
	uint16_t tlv_len = tlv_length(tlvs[0][0]);

	t = (uint8_t *)tlvs[0][0]->data;
	if (!t)
		return false;

	/* offset within provided minlen */
	offset += 1; /* opclass */
	num_sta = t[offset];
	offset += 1; /* num sta */

	if ((offset + num_sta * 12) > tlv_len)
		return false;

	return true;
}

bool validate_channel_preference_report(struct cmdu_buff *cmdu,
					struct tlv *tlvs[][TLV_MAXNUM], size_t tlvs_size,
					uint8_t map_profile)
{
	int index;
	int ret = map_cmdu_parse_tlvs(cmdu, tlvs, tlvs_size, map_profile);


	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));
	}

	/* Channel Preference TLV */
	index = 0;
	while (index < TLV_MAXNUM && tlvs[0][index]) {
		int i, offset = 0;
		int num_opclass = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tlvs[0][index]);

		t = (uint8_t *)tlvs[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* radio id */
		num_opclass = t[offset];
		offset += 1;	/* num opclass */

		for (i = 0; i < num_opclass; i++) {
			int num_ch = 0;

			if ((offset + 3) > tlv_len)
				return false;

			offset += 1;	/* class id */
			num_ch = t[offset];
			offset += 1;	/* num channel */
			if ((offset + num_ch + 1) > tlv_len)
				return false;

			offset += num_ch;	/* channel */
			offset += 1;	/* preference */
		}

		index++;
	}


	/* CAC Status Report TLV */
	if (tlvs[3][0]) {
		const uint16_t tlv_len = tlv_length(tlvs[3][0]);
		const uint8_t *tlv_data = (uint8_t *)tlvs[3][0]->data;
		const struct cac_status_available *avail_channels =
			(struct cac_status_available *)tlv_data;
		const struct cac_status_noop *noop_channels = NULL;
		const struct cac_status_cac *active_channels = NULL;

		int offset = sizeof(avail_channels->num_channel) +
			     avail_channels->num_channel * sizeof(avail_channels->ch[0]);

		if (offset > tlv_len)
			return false;

		noop_channels = (struct cac_status_noop *)(tlv_data + offset);
		offset += sizeof(noop_channels->num_channel) +
			  noop_channels->num_channel * sizeof(noop_channels->ch[0]);

		if (offset > tlv_len)
			return false;

		active_channels = (struct cac_status_cac *)(tlv_data + offset);
		offset += sizeof(active_channels->num_channel) +
			  active_channels->num_channel * sizeof(active_channels->ch[0]);
		if (offset > tlv_len)
			return false;
	}

	return true;
}
