
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include "timer.h"
#include <easy/easy.h>
#include <wifiutils.h>
#include <cmdu.h>

#include "utils.h"
#include "wifi_dataelements.h"
#include "decollector.h"

static int cntlr_config_get_ssid(struct uci_section *s,
				 struct list_head *ssidlist,
				 uint32_t *num)
{
	enum {
		SSID_BAND,
		SSID_SSID,
		SSID_SEC,
		SSID_KEY,
		SSID_VLAN,
		SSID_NETWORK,
		SSID_TYPE,
		SSID_DISALLOW_BSTA,
		SSID_ENABLED,
		NUM_SSID_ATTRS,
	};
	const struct uci_parse_option opts[] = {
		[SSID_BAND] = { .name = "band", .type = UCI_TYPE_STRING },
		[SSID_SSID] = { .name = "ssid", .type = UCI_TYPE_STRING },
		[SSID_SEC] = { .name = "encryption", .type = UCI_TYPE_STRING },
		[SSID_KEY] = { .name = "key", .type = UCI_TYPE_STRING },
		[SSID_VLAN] = { .name = "vid", .type = UCI_TYPE_STRING },
		[SSID_NETWORK] = { .name = "network", .type = UCI_TYPE_STRING },
		[SSID_TYPE] = { .name = "type", .type = UCI_TYPE_STRING },
		[SSID_DISALLOW_BSTA] = { .name = "disallow_bsta", .type = UCI_TYPE_LIST },
		[SSID_ENABLED] = { .name = "enabled", .type = UCI_TYPE_STRING }
	};
	struct uci_option *tb[NUM_SSID_ATTRS];
	struct wifi_network_ssid *ssid;


	ssid = calloc(1, sizeof(*ssid));
	if (!ssid)
		return -1;

	ssid->enabled = true;
	ssid->band = BAND_UNKNOWN;
	ssid->type = AP_WIFI_FBSS;
	ssid->multi_ap = 2;

	uci_parse_section(s, opts, NUM_SSID_ATTRS, tb);

	if (tb[SSID_BAND]) {
		if (atoi(tb[SSID_BAND]->v.string) == 5)
			ssid->band = BAND_5;
		else if (atoi(tb[SSID_BAND]->v.string) == 2)
			ssid->band = BAND_2;
	}

	if (tb[SSID_SSID])
		strncpy((char *)ssid->ssid, tb[SSID_SSID]->v.string, 32);

	if (tb[SSID_SEC]) {
		const char *sec = tb[SSID_SEC]->v.string;

		if (!strncmp(sec, "sae-mixed", 9)) {
			ssid->security |= BIT(WIFI_SECURITY_WPA3PSK);
			ssid->security |= BIT(WIFI_SECURITY_WPA3PSK_T);
		} else if (!strncmp(sec, "sae", 3)) {
			ssid->security |= BIT(WIFI_SECURITY_WPA3PSK);
		} else if (!strncmp(sec, "psk-mixed", 9)) {
			ssid->security |= BIT(WIFI_SECURITY_WPAPSK);
			ssid->security |= BIT(WIFI_SECURITY_WPA2PSK);
		} else if (!strncmp(sec, "psk2", 4)) {
			ssid->security |= BIT(WIFI_SECURITY_WPA2PSK);
		} else if (!strncmp(sec, "psk", 3)) {
			ssid->security |= BIT(WIFI_SECURITY_WPAPSK);
		} else if (!strncmp(sec, "wpa-mixed", 9)) {
			ssid->security |= BIT(WIFI_SECURITY_WPA);
			ssid->security |= BIT(WIFI_SECURITY_WPA2);
		} else if (!strncmp(sec, "wpa2", 4)) {
			ssid->security |= BIT(WIFI_SECURITY_WPA2);
		} else if (!strncmp(sec, "wpa", 3)) {
			ssid->security |= BIT(WIFI_SECURITY_WPA);
		} else if (!strncmp(sec, "none", 4)) {
			ssid->security |= BIT(WIFI_SECURITY_NONE);
		} else if (!strncmp(sec, "open", 4)) {
			ssid->security |= BIT(WIFI_SECURITY_NONE);
		} else {
			free(ssid);
			return -1;
		}
	}

	if (tb[SSID_VLAN])
		ssid->vid = (uint16_t) atoi(tb[SSID_VLAN]->v.string);

	if (tb[SSID_TYPE]) {
		const char *type = tb[SSID_TYPE]->v.string;

		if (!strcmp(type, "backhaul")) {
			ssid->multi_ap = 1;
			ssid->type = AP_WIFI_BBSS;
		} else if (!strcmp(type, "fronthaul")) {
			ssid->multi_ap = 2;
			ssid->type = AP_WIFI_FBSS;
		} else if (!strcmp(type, "combined")) {
			ssid->multi_ap = 3;
			ssid->type = AP_WIFI_COMBINED;
		} else {
			free(ssid);
			return -1;
		}
	}

	if (tb[SSID_ENABLED])
		ssid->enabled = atoi(tb[SSID_ENABLED]->v.string);

	*num += 1;
	list_add_tail(&ssid->list, ssidlist);
	return 0;
}

static int cntlr_config_get_default8021q(struct uci_section *s,
					 struct wifi_network *net)
{
	enum {
		NODE_PVID,
		NODE_PCP,
		NUM_POLICIES,
	};
	const struct uci_parse_option opts[] = {
		{ .name = "primary_vid", .type = UCI_TYPE_STRING },
		{ .name = "primary_pcp", .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_POLICIES];

	uci_parse_section(s, opts, NUM_POLICIES, tb);

	if (tb[NODE_PVID])
		net->primary_vid = atoi(tb[NODE_PVID]->v.string);

	if (tb[NODE_PCP])
		net->default_pcp = atoi(tb[NODE_PCP]->v.string);

	return 0;
}

static int cntlr_config_get_agent_node(struct uci_section *s,
				       struct wifi_network *net)
{
	enum {
		NODE_AGENT_ID,
		NODE_RPT_ASSOC_FAILS,
		NODE_RPT_ASSOC_FAILS_RATE,
		NODE_RPT_METRIC_PERIODIC,
		NODE_RPT_SCAN,
		NODE_STEER_EXCLUDE,
		NODE_STEER_EXCLUDE_BTM,
		NODE_STA_STEER,
		NUM_POLICIES,
	};
	const struct uci_parse_option opts[] = {
		[NODE_AGENT_ID] =
			{ .name = "agent_id", .type = UCI_TYPE_STRING },
		[NODE_RPT_ASSOC_FAILS] =
			{ .name = "report_sta_assocfails", .type = UCI_TYPE_STRING },
		[NODE_RPT_ASSOC_FAILS_RATE] =
			{ .name = "report_sta_assocfails_rate", .type = UCI_TYPE_STRING },
		[NODE_RPT_METRIC_PERIODIC] =
			{ .name = "report_metric_periodic", .type = UCI_TYPE_STRING },
		[NODE_RPT_SCAN] =
			{ .name = "report_scan", .type = UCI_TYPE_STRING },
		[NODE_STEER_EXCLUDE] =
			{ .name = "steer_exclude", .type = UCI_TYPE_LIST },
		[NODE_STEER_EXCLUDE_BTM] =
			{ .name = "steer_exclude_btm", .type = UCI_TYPE_LIST },
		[NODE_STA_STEER] =
			{ .name = "sta_steer", .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_POLICIES];
	struct wifi_network_device *dev = NULL;
	bool device_found = false;

	uci_parse_section(s, opts, NUM_POLICIES, tb);

	if (tb[NODE_AGENT_ID]) {
		macaddr_t agent_id;

		hwaddr_aton(tb[NODE_AGENT_ID]->v.string, agent_id);

		list_for_each_entry(dev, &net->devicelist, list) {
			if (!memcmp(dev->macaddr, agent_id, 6)) {
				device_found = true;
				break;
			}
		}
	} else {
		return -1;
	}

	/* No device in model yet */
	if (!device_found)
		return 0;

	if (tb[NODE_RPT_ASSOC_FAILS]) {
		dev->report.sta_assocfails =
			atoi(tb[NODE_RPT_ASSOC_FAILS]->v.string) == 1 ? true : false;
	}

	if (tb[NODE_RPT_ASSOC_FAILS_RATE]) {
		dev->report.sta_assocfails_rate =
				atoi(tb[NODE_RPT_ASSOC_FAILS_RATE]->v.string);
	}

	if (tb[NODE_RPT_SCAN]) {
		dev->report.independent_scans =
			atoi(tb[NODE_RPT_SCAN]->v.string) == 1 ? true : false;
	}

	if (tb[NODE_RPT_METRIC_PERIODIC]) {
		dev->report.ap_metrics_int =
			atoi(tb[NODE_RPT_METRIC_PERIODIC]->v.string);
	}

	if (tb[NODE_STEER_EXCLUDE]) {
		struct uci_element *excluded_sta;

		list_flush(&dev->sta_steer_disallowlist, struct mac_address_element, list);

		uci_foreach_element(&tb[NODE_STEER_EXCLUDE]->v.list, excluded_sta) {
			struct mac_address_element *mac =
				calloc(1, sizeof(*mac));

			if (mac && hwaddr_aton(excluded_sta->name, mac->macaddr))
				list_add_tail(&mac->list, &dev->sta_steer_disallowlist);
		}
	}

	if (tb[NODE_STEER_EXCLUDE_BTM]) {
		struct uci_element *excluded_sta;

		list_flush(&dev->sta_btmsteer_disallowlist, struct mac_address_element, list);

		uci_foreach_element(&tb[NODE_STEER_EXCLUDE_BTM]->v.list, excluded_sta) {
			struct mac_address_element *mac =
				calloc(1, sizeof(*mac));

			if (mac && hwaddr_aton(excluded_sta->name, mac->macaddr))
				list_add_tail(&mac->list, &dev->sta_btmsteer_disallowlist);
		}
	}

	if (tb[NODE_STA_STEER]) {
		dev->sta_steering_state =
			atoi(tb[NODE_STA_STEER]->v.string) == 1 ? true : false;
	}

	return 0;
}

int collector_get_network_params(struct decollector_private *priv,
				 struct wifi_network *net)
{
	struct list_head *ssidlist = &net->ssidlist;
	uint32_t *num = &net->num_ssid;
	struct wifi_network_ssid *ssid, *tssid;

	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, "mapcontroller", &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	*num = 0;
	net->num_ssid = 0;
	list_for_each_entry_safe(ssid, tssid, &net->ssidlist, list) {
		list_del(&ssid->list);
		free(ssid);
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "ap"))
			cntlr_config_get_ssid(s, ssidlist, num);
		else if (!strcmp(s->type, "controller"))
			cntlr_config_get_default8021q(s, net);
		else if (!strcmp(s->type, "node"))
			cntlr_config_get_agent_node(s, net);
	}

	uci_free_context(ctx);
	return 0;
}
