/*
 * decollector.h
 * WiFi DataElements Collector private header file.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#ifndef DECOLLECTOR_H
#define DECOLLECTOR_H

#include "backhaul_topology.h"

#include <libubox/list.h>
#include <stdint.h>

#define WIFI_DEVICE_MAX_NUM	16

enum {
	WIFI_IFACE_FH,
	WIFI_IFACE_BK,
};

#define MAP_PROFILE_1	1
#define MAP_PROFILE_2	2

typedef uint32_t object_t;

#define OBJECT_INVALID 0

#define IEEE1905_OBJECT			"ieee1905"
#define MAP_PLUGIN_OBJECT		"ieee1905.map"
#define WFADATAD_AGENT_OBJECT		"wifi.dataelements"
#define WFADATAD_COLLECTOR_OBJECT	"wifi.dataelements.collector"
#define MAP_CONTROLLER_OBJECT		"map.controller"

struct wifi_assoc_frame {
	uint16_t len;
	uint8_t macaddr[6];
	uint8_t *frame;
	struct list_head list;
};

struct decollector_config {
	uint32_t refresh_int;
	bool is_collector;
	bool is_sap_mode;
	uint32_t profile;
	struct ubus_context *uctx;
	char mapcontroller_conffile[256];
};

#define DELM_OBJECT_PLACEHOLDER			\
	struct ubus_context *ctx;		\
	struct ubus_object obj;			\
	struct ubus_event_handler topology_ev;	\
	struct ubus_event_handler ubus_obj_ev;

enum collection_state {
	CT_IDLE = 0,
	CT_RUNNING,
	CT_DONE,
};

struct decollector_private {
	struct bh_topology_dev_list bh_topo_dev_list;
	struct decollector_config cfg;
	time_t refresh_time;
	atimer_t t;
	uint8_t cstate;			/* collector state */
	uint8_t cntlr_id[6];		/* controller id */
	bool dm_available;
	struct wifi_data_element *dm;
	int pending;
	object_t i1905;			/* i1905 UBUS object */
	object_t i1905_map;		/* i1905.map UBUS object */
	uint8_t ieee1905_macaddr[6];	/* self node macaddr */
	void *subscriber;
	struct list_head framelist;	/* wifi_assoc_frame list */
	DELM_OBJECT_PLACEHOLDER
};

#define CMDU_TYPE_INVALID	-1


extern int decollector_publish_object(struct decollector_private *p, const char *objname);
extern int decollector_register_events(struct decollector_private *p);
extern int decollector_subscribe_for_cmdus(struct decollector_private *priv);
extern uint32_t decollector_lookup_object(struct decollector_private *p, const char *objname);


extern int decollector_dump_all(struct decollector_private *p, struct blob_buf *bb);
extern int decollector_dump2_all(struct decollector_private *p, struct blob_buf *bb);
extern int decollector_event_dump(struct decollector_private *p, struct blob_buf *bb);
extern int decollector_notify_events(struct decollector_private *p, struct wifi_network_device *dev);
extern int decollector_collect_node(struct decollector_private *p, struct wifi_network_device *dev);
extern struct wifi_network_device *decollector_get_origin_dev(
		struct decollector_private *p, struct wifi_data_element *dm,
		uint8_t *origin);

extern int decollector_send_cmdu_request(struct decollector_private *priv,
					 struct cmdu_buff *req, uint8_t *dst);

extern int decollector_get_deagents(struct decollector_private *p, int *num, uint8_t **macaddrs);
extern int decollector_get_ieee1905id(struct decollector_private *p);
extern int decollector_get_network_steer_stats(struct decollector_private *p);
extern int decollector_get_sta_steer_stats(struct decollector_private *p);


extern int decollector_alloc_deagents(struct decollector_private *priv);
extern void decollector_getter_run(void *data);
extern void decollector_getter_run_init(struct decollector_private *p);
extern void decollector_refresh_all(struct decollector_private *priv);

extern void decollector_free_dm(struct wifi_data_element *dm);
extern void free_network_device(struct wifi_network_device *dev);
extern void free_sta(struct wifi_sta_element *sta);
extern void free_bss(struct wifi_bss_element *bss);
extern void free_radio(struct wifi_radio_element *r);
extern void free_scanresult_data(struct wifi_scanres_element *scanres);
extern void free_wifi_cac_status(struct wifi_cac_status *cac_status);
extern void clear_supported_opclasses(struct wifi_radio_element *radio);
extern void clear_current_opclasses(struct wifi_radio_element *radio);

extern int prepare_autoconfig_search(struct decollector_private *priv, uint8_t band);


extern int decollector_alloc_getter(struct decollector_private *p, struct wifi_network_device *dev);
extern int decollector_free_getters(struct decollector_private *p);
extern int decollector_free_getter(struct wifi_network_device *dev);
extern int decollector_sched_getter(struct decollector_private *p, struct wifi_network_device *dev,
				    uint16_t msg, uint32_t after_ms);

extern int decollector_stop_getter(struct decollector_private *p, struct wifi_network_device *dev);
extern void decollector_stop_getter_all(struct decollector_private *p);
extern int decollector_getter_running(struct decollector_private *p, struct wifi_network_device *dev);

extern int decollector_getter_get_state(struct decollector_private *p,
		struct wifi_network_device *dev, uint8_t *state);
extern int decollector_getter_set_state(struct decollector_private *p,
		struct wifi_network_device *dev, uint8_t state);

extern int prepare_1905_query(struct decollector_private *priv, uint16_t cmdu_type,
				struct wifi_network_device *dev, uint8_t *dst);

#endif /* DECOLLECTOR_H */
